"""Tests for agent-level prompt caching in Thread._apply_prompt_caching."""
import pytest
from localrouter import (
    ChatMessage,
    MessageRole,
    TextBlock,
    ToolDefinition,
    CacheControl,
)

# Import Thread from automator
import sys
sys.path.insert(0, 'packages/automator/src')
from automator.agent import Thread


class TestThreadPromptCaching:
    """Tests for Thread._apply_prompt_caching method."""
    
    def test_caching_not_applied_to_non_claude_models(self):
        """Test that caching is not applied to non-Claude models."""
        thread = Thread(llm={"model": "gpt-4"}, messages=[], tools=[])
        
        messages = [
            ChatMessage(
                role=MessageRole.system,
                content=[TextBlock(text="You are helpful")]
            ),
            ChatMessage(
                role=MessageRole.user,
                content=[TextBlock(text="Hello")]
            )
        ]
        tools = [
            ToolDefinition(
                name="test_tool",
                description="A test tool",
                input_schema={"type": "object", "properties": {}}
            )
        ]
        
        cached_msgs, cached_tools = thread._apply_prompt_caching(messages, tools, "gpt-4")
        
        # No cache_control should be added
        assert cached_msgs[0].content[0].cache_control is None
        assert cached_msgs[1].content[0].cache_control is None
        assert cached_tools[0].cache_control is None
    
    def test_caching_applied_to_claude_models(self):
        """Test that caching is applied to Claude models."""
        thread = Thread(llm={"model": "claude-sonnet-4-5-20250929"}, messages=[], tools=[])
        
        messages = [
            ChatMessage(
                role=MessageRole.system,
                content=[TextBlock(text="You are helpful")]
            ),
            ChatMessage(
                role=MessageRole.user,
                content=[TextBlock(text="Hello")]
            )
        ]
        tools = [
            ToolDefinition(
                name="test_tool",
                description="A test tool",
                input_schema={"type": "object", "properties": {}}
            )
        ]
        
        cached_msgs, cached_tools = thread._apply_prompt_caching(messages, tools, "claude-sonnet-4-5-20250929")
        
        # System message last block should have cache_control
        assert cached_msgs[0].content[-1].cache_control is not None
        assert cached_msgs[0].content[-1].cache_control.type == "ephemeral"
        
        # Last tool should have cache_control
        assert cached_tools[-1].cache_control is not None
        assert cached_tools[-1].cache_control.type == "ephemeral"
        
        # Last message last block should have cache_control
        assert cached_msgs[-1].content[-1].cache_control is not None
        assert cached_msgs[-1].content[-1].cache_control.type == "ephemeral"
    
    def test_caching_does_not_mutate_originals(self):
        """Test that caching creates copies and doesn't mutate original messages."""
        thread = Thread(llm={"model": "claude-sonnet-4-5-20250929"}, messages=[], tools=[])
        
        original_text_block = TextBlock(text="Hello")
        messages = [
            ChatMessage(
                role=MessageRole.user,
                content=[original_text_block]
            )
        ]
        original_tool = ToolDefinition(
            name="test_tool",
            description="A test tool",
            input_schema={"type": "object", "properties": {}}
        )
        tools = [original_tool]
        
        cached_msgs, cached_tools = thread._apply_prompt_caching(messages, tools, "claude-sonnet-4-5-20250929")
        
        # Original should not be modified
        assert original_text_block.cache_control is None
        assert original_tool.cache_control is None
        
        # Cached versions should have cache_control
        assert cached_msgs[0].content[0].cache_control is not None
        assert cached_tools[0].cache_control is not None
    
    def test_caching_with_multiple_system_blocks(self):
        """Test caching when system message has multiple blocks."""
        thread = Thread(llm={"model": "claude-sonnet-4-5-20250929"}, messages=[], tools=[])
        
        messages = [
            ChatMessage(
                role=MessageRole.system,
                content=[
                    TextBlock(text="You are helpful."),
                    TextBlock(text="<context>Important info</context>"),
                    TextBlock(text="Follow these rules.")
                ]
            ),
            ChatMessage(
                role=MessageRole.user,
                content=[TextBlock(text="Hello")]
            )
        ]
        
        cached_msgs, _ = thread._apply_prompt_caching(messages, [], "claude-sonnet-4-5-20250929")
        
        # Only the LAST block of system message should have cache_control
        assert cached_msgs[0].content[0].cache_control is None
        assert cached_msgs[0].content[1].cache_control is None
        assert cached_msgs[0].content[2].cache_control is not None
        assert cached_msgs[0].content[2].cache_control.type == "ephemeral"
    
    def test_caching_with_multiple_tools(self):
        """Test caching when there are multiple tools."""
        thread = Thread(llm={"model": "claude-sonnet-4-5-20250929"}, messages=[], tools=[])
        
        messages = [
            ChatMessage(
                role=MessageRole.user,
                content=[TextBlock(text="Hello")]
            )
        ]
        tools = [
            ToolDefinition(
                name="tool1",
                description="First tool",
                input_schema={"type": "object", "properties": {}}
            ),
            ToolDefinition(
                name="tool2",
                description="Second tool",
                input_schema={"type": "object", "properties": {}}
            ),
            ToolDefinition(
                name="tool3",
                description="Third tool",
                input_schema={"type": "object", "properties": {}}
            )
        ]
        
        _, cached_tools = thread._apply_prompt_caching(messages, tools, "claude-sonnet-4-5-20250929")
        
        # Only the LAST tool should have cache_control
        assert cached_tools[0].cache_control is None
        assert cached_tools[1].cache_control is None
        assert cached_tools[2].cache_control is not None
        assert cached_tools[2].cache_control.type == "ephemeral"
    
    def test_caching_with_no_system_message(self):
        """Test caching when there's no system message."""
        thread = Thread(llm={"model": "claude-sonnet-4-5-20250929"}, messages=[], tools=[])
        
        messages = [
            ChatMessage(
                role=MessageRole.user,
                content=[TextBlock(text="Hello")]
            ),
            ChatMessage(
                role=MessageRole.assistant,
                content=[TextBlock(text="Hi there!")]
            ),
            ChatMessage(
                role=MessageRole.user,
                content=[TextBlock(text="How are you?")]
            )
        ]
        
        cached_msgs, _ = thread._apply_prompt_caching(messages, [], "claude-sonnet-4-5-20250929")
        
        # Last message's last block should have cache_control
        assert cached_msgs[-1].content[-1].cache_control is not None
        
        # Other messages should not have cache_control
        assert cached_msgs[0].content[0].cache_control is None
        assert cached_msgs[1].content[0].cache_control is None
    
    def test_caching_with_empty_tools(self):
        """Test caching when tools list is empty."""
        thread = Thread(llm={"model": "claude-sonnet-4-5-20250929"}, messages=[], tools=[])
        
        messages = [
            ChatMessage(
                role=MessageRole.user,
                content=[TextBlock(text="Hello")]
            )
        ]
        
        cached_msgs, cached_tools = thread._apply_prompt_caching(messages, [], "claude-sonnet-4-5-20250929")
        
        # Should not error and should still cache messages
        assert cached_tools == []
        assert cached_msgs[-1].content[-1].cache_control is not None


"""Tests for Anthropic prompt caching support."""
import pytest
from localrouter import (
    ChatMessage,
    MessageRole,
    TextBlock,
    ImageBlock,
    ToolUseBlock,
    ToolResultBlock,
    ToolDefinition,
    CacheControl,
    anthropic_format,
)


class TestCacheControlModel:
    """Tests for the CacheControl model."""
    
    def test_cache_control_default(self):
        """Test that CacheControl defaults to ephemeral type."""
        cc = CacheControl()
        assert cc.type == "ephemeral"
    
    def test_cache_control_model_dump(self):
        """Test CacheControl serialization."""
        cc = CacheControl(type="ephemeral")
        assert cc.model_dump() == {"type": "ephemeral"}


class TestContentBlockCaching:
    """Tests for cache_control on content blocks."""
    
    def test_text_block_with_cache_control(self):
        """Test TextBlock with cache_control in anthropic_format."""
        block = TextBlock(
            text="Hello world",
            cache_control=CacheControl(type="ephemeral")
        )
        result = block.anthropic_format()
        assert result == {
            "type": "text",
            "text": "Hello world",
            "cache_control": {"type": "ephemeral"}
        }
    
    def test_text_block_without_cache_control(self):
        """Test TextBlock without cache_control."""
        block = TextBlock(text="Hello world")
        result = block.anthropic_format()
        assert result == {"type": "text", "text": "Hello world"}
        assert "cache_control" not in result
    
    def test_tool_use_block_with_cache_control(self):
        """Test ToolUseBlock with cache_control."""
        block = ToolUseBlock(
            id="test-123",
            name="test_tool",
            input={"query": "test"},
            cache_control=CacheControl(type="ephemeral")
        )
        result = block.anthropic_format()
        assert result["cache_control"] == {"type": "ephemeral"}
    
    def test_tool_result_block_with_cache_control(self):
        """Test ToolResultBlock with cache_control."""
        block = ToolResultBlock(
            tool_use_id="test-123",
            content=[TextBlock(text="Result")],
            cache_control=CacheControl(type="ephemeral")
        )
        result = block.anthropic_format()
        assert result["cache_control"] == {"type": "ephemeral"}


class TestToolDefinitionCaching:
    """Tests for cache_control on tool definitions."""
    
    def test_tool_definition_with_cache_control(self):
        """Test ToolDefinition with cache_control in anthropic_format."""
        tool = ToolDefinition(
            name="test_tool",
            description="A test tool",
            input_schema={"type": "object", "properties": {}},
            cache_control=CacheControl(type="ephemeral")
        )
        result = tool.anthropic_format
        assert result["cache_control"] == {"type": "ephemeral"}
    
    def test_tool_definition_without_cache_control(self):
        """Test ToolDefinition without cache_control."""
        tool = ToolDefinition(
            name="test_tool",
            description="A test tool",
            input_schema={"type": "object", "properties": {}}
        )
        result = tool.anthropic_format
        assert "cache_control" not in result


class TestAnthropicFormatCaching:
    """Tests for anthropic_format function with caching."""
    
    def test_system_message_with_cache_control_uses_structured_format(self):
        """Test that system messages with cache_control use structured format."""
        system_msg = ChatMessage(
            role=MessageRole.system,
            content=[
                TextBlock(text="You are a helpful assistant."),
                TextBlock(
                    text="<context>Important context here</context>",
                    cache_control=CacheControl(type="ephemeral")
                )
            ]
        )
        user_msg = ChatMessage(
            role=MessageRole.user,
            content=[TextBlock(text="Hello")]
        )
        
        result = anthropic_format([system_msg, user_msg], tools=None)
        
        # Should use structured system format (list of dicts)
        assert isinstance(result["system"], list)
        assert len(result["system"]) == 2
        
        # First block should not have cache_control
        assert result["system"][0]["type"] == "text"
        assert "cache_control" not in result["system"][0]
        
        # Second block should have cache_control
        assert result["system"][1]["type"] == "text"
        assert result["system"][1]["cache_control"] == {"type": "ephemeral"}
    
    def test_system_message_without_cache_control_uses_string_format(self):
        """Test that system messages without cache_control use string format."""
        system_msg = ChatMessage(
            role=MessageRole.system,
            content=[TextBlock(text="You are a helpful assistant.")]
        )
        user_msg = ChatMessage(
            role=MessageRole.user,
            content=[TextBlock(text="Hello")]
        )
        
        result = anthropic_format([system_msg, user_msg], tools=None)
        
        # Should use simple string format
        assert isinstance(result["system"], str)
        assert result["system"] == "You are a helpful assistant."
    
    def test_tools_with_cache_control(self):
        """Test that tools with cache_control are formatted correctly."""
        tools = [
            ToolDefinition(
                name="tool1",
                description="First tool",
                input_schema={"type": "object", "properties": {}}
            ),
            ToolDefinition(
                name="tool2",
                description="Second tool",
                input_schema={"type": "object", "properties": {}},
                cache_control=CacheControl(type="ephemeral")
            )
        ]
        
        messages = [
            ChatMessage(
                role=MessageRole.user,
                content=[TextBlock(text="Hello")]
            )
        ]
        
        result = anthropic_format(messages, tools=tools)
        
        # First tool should not have cache_control
        assert "cache_control" not in result["tools"][0]
        
        # Last tool should have cache_control
        assert result["tools"][1]["cache_control"] == {"type": "ephemeral"}
    
    def test_message_content_with_cache_control(self):
        """Test that message content with cache_control is preserved."""
        messages = [
            ChatMessage(
                role=MessageRole.user,
                content=[
                    TextBlock(text="First part"),
                    TextBlock(
                        text="Cached part",
                        cache_control=CacheControl(type="ephemeral")
                    )
                ]
            )
        ]
        
        result = anthropic_format(messages, tools=None)
        
        # Message content should have cache_control on last block
        msg_content = result["messages"][0]["content"]
        assert len(msg_content) == 2
        assert "cache_control" not in msg_content[0]
        assert msg_content[1]["cache_control"] == {"type": "ephemeral"}


class TestCacheControlSerialization:
    """Tests for cache_control serialization and deserialization."""
    
    def test_text_block_roundtrip(self):
        """Test that TextBlock with cache_control survives model_dump/reconstruct."""
        original = TextBlock(
            text="Hello",
            cache_control=CacheControl(type="ephemeral")
        )
        
        dumped = original.model_dump()
        reconstructed = TextBlock(**dumped)
        
        assert reconstructed.cache_control is not None
        assert reconstructed.cache_control.type == "ephemeral"
    
    def test_chat_message_roundtrip(self):
        """Test that ChatMessage with cached blocks survives roundtrip."""
        original = ChatMessage(
            role=MessageRole.user,
            content=[
                TextBlock(
                    text="Cached content",
                    cache_control=CacheControl(type="ephemeral")
                )
            ]
        )
        
        dumped = original.model_dump()
        reconstructed = ChatMessage(**dumped)
        
        assert len(reconstructed.content) == 1
        assert reconstructed.content[0].cache_control is not None
        assert reconstructed.content[0].cache_control.type == "ephemeral"